/*
 * Copyright (C) 2000-2001 Michael Smith (msmith at xiph org)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, version 2.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 */

#include <stdio.h>
#include <stdbool.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <limits.h>
#include <unistd.h>
#include <ogg/ogg.h>
#include <vorbis/codec.h>
#include <assert.h>

#include "vcedit.h"

#define CHUNKSIZE 4096

struct vcedit_state_St {
	int refcount;

	ogg_sync_state oy;
	ogg_stream_state os;

	vorbis_comment vc;
	vorbis_info vi;

	FILE *in;
	bool opened;
	long serial;
	unsigned char *mainbuf;
	unsigned char *bookbuf;
	int	mainlen;
	int	booklen;
	char *vendor;
	int prevW;
	int extrapage;
	int eosin;

	char filename[0];
};

static void
vcedit_state_free (vcedit_state *s)
{
	free (s->mainbuf);
	free (s->bookbuf);
	free (s->vendor);

	if (s->in) {
		fclose (s->in);
		s->in = NULL;
	}

	free (s);
}

static bool
vcedit_state_init (vcedit_state *s, const char *filename)
{
	s->refcount = 1;

	strcpy (s->filename, filename);

	return true;
}

vcedit_state *
vcedit_state_new (const char *filename)
{
	vcedit_state *s;
	size_t len;

	len = strlen (filename);
	if (len > PATH_MAX)
		return NULL;

	s= malloc (sizeof (vcedit_state) + len + 1);
	if (!s)
		return NULL;

	memset (s, 0, sizeof (vcedit_state));

	if (!vcedit_state_init (s, filename)) {
		vcedit_state_free (s);
		return NULL;
	}

	return s;
}

vorbis_comment *
vcedit_comments (vcedit_state *s)
{
	return s->opened ? &s->vc : NULL;
}

static void
vcedit_clear_internals (vcedit_state *s)
{
	ogg_stream_clear (&s->os);
	ogg_sync_clear (&s->oy);

	vorbis_info_clear (&s->vi);
	vorbis_comment_clear (&s->vc);

	free (s->vendor);
	s->vendor = NULL;

	free (s->mainbuf);
	s->mainbuf = NULL;
	s->mainlen = 0;

	free (s->bookbuf);
	s->bookbuf = NULL;
	s->booklen = 0;

	s->serial = 0;
	s->opened = false;
}

void
vcedit_state_ref (vcedit_state *s)
{
	s->refcount++;
}

void
vcedit_state_unref (vcedit_state *s)
{
	if (--s->refcount)
		return;

	if (s->opened)
		vcedit_clear_internals (s);

	vcedit_state_free (s);
}

/* Next two functions pulled straight from libvorbis, apart from one change
 * - we don't want to overwrite the vendor string.
 */
static void
_v_writestring (oggpack_buffer *o, char *s, int len)
{
	while (len--)
		oggpack_write (o, *s++, 8);
}

static int
_commentheader_out (vorbis_comment *vc, char *vendor, ogg_packet *op)
{
	int i;

	oggpack_buffer opb;

	oggpack_writeinit (&opb);

	/* preamble */
	oggpack_write (&opb, 0x03, 8);
	_v_writestring (&opb, "vorbis", 6);

	/* vendor */
	oggpack_write (&opb, strlen (vendor), 32);
	_v_writestring (&opb, vendor, strlen (vendor));

	/* comments */
	oggpack_write (&opb, vc->comments, 32);

	for (i = 0; i < vc->comments; i++) {
		if (!vc->user_comments[i])
			oggpack_write (&opb, 0, 32);
		else {
			oggpack_write (&opb, vc->comment_lengths[i], 32);
			_v_writestring (&opb, vc->user_comments[i],
			                vc->comment_lengths[i]);
		}
	}

	oggpack_write (&opb, 1, 1);

	op->packet = _ogg_malloc (oggpack_bytes (&opb));
	memcpy (op->packet, opb.buffer, oggpack_bytes (&opb));

	op->bytes = oggpack_bytes (&opb);
	op->b_o_s = 0;
	op->e_o_s = 0;
	op->granulepos = 0;

	oggpack_writeclear (&opb);

	return 0;
}

static int
_blocksize (vcedit_state *s, ogg_packet *p)
{
	int this, ret = 0;

	this = vorbis_packet_blocksize (&s->vi, p);

	if (s->prevW)
		ret = (this + s->prevW) / 4;

	s->prevW = this;

	return ret;
}

static int
_fetch_next_packet (vcedit_state *s, ogg_packet *p, ogg_page *page)
{
	char *buffer;
	int result, bytes;

	result = ogg_stream_packetout (&s->os, p);
	if (result == 1)
		return 1;

	if (s->eosin)
		return 0;

	while (ogg_sync_pageout (&s->oy, page) != 1) {
		buffer = ogg_sync_buffer (&s->oy, CHUNKSIZE);
		bytes = fread (buffer, 1, CHUNKSIZE, s->in);
		ogg_sync_wrote (&s->oy, bytes);

		if (!bytes && (feof (s->in) || ferror (s->in)))
			return 0;
	}

	if (ogg_page_eos (page))
		s->eosin = 1;
	else if (ogg_page_serialno (page) != s->serial) {
		s->eosin = 1;
		s->extrapage = 1;
		return 0;
	}

	ogg_stream_pagein (&s->os, page);

	return _fetch_next_packet (s, p, page);
}

vcedit_error
vcedit_open (vcedit_state *s)
{
	vcedit_error ret;
	ogg_packet *header;
	ogg_packet header_main, header_comments, header_codebooks;
	ogg_page og;
	char *buffer;
	size_t bytes, total = 0;
	int i = 0;

	s->in = fopen (s->filename, "rb");
	if (!s->in)
		return VCEDIT_ERR_OPEN;

	ogg_sync_init (&s->oy);

	do {
		/* Bail if we don't find data in the first 40 kB */
		if (feof (s->in) || ferror (s->in) || total >= (CHUNKSIZE * 10)) {
			ogg_sync_clear (&s->oy);

			return VCEDIT_ERR_INVAL;
		}

		buffer = ogg_sync_buffer (&s->oy, CHUNKSIZE);

		bytes = fread (buffer, 1, CHUNKSIZE, s->in);
		total += bytes;

		ogg_sync_wrote (&s->oy, bytes);
	} while (ogg_sync_pageout (&s->oy, &og) != 1);

	s->serial = ogg_page_serialno (&og);

	ogg_stream_init (&s->os, s->serial);
	vorbis_info_init (&s->vi);
	vorbis_comment_init (&s->vc);

	if (ogg_stream_pagein (&s->os, &og) < 0) {
		ret = VCEDIT_ERR_INVAL;
		goto err;
	}

	if (ogg_stream_packetout (&s->os, &header_main) != 1) {
		ret = VCEDIT_ERR_INVAL;
		goto err;
	}

	if (vorbis_synthesis_headerin (&s->vi, &s->vc, &header_main) < 0) {
		ret = VCEDIT_ERR_INVAL;
		goto err;
	}

	s->mainlen = header_main.bytes;
	s->mainbuf = malloc (s->mainlen);
	memcpy (s->mainbuf, header_main.packet, header_main.bytes);

	header = &header_comments;

	while (i < 2) {
		if (feof (s->in) || ferror (s->in)) {
			ret = VCEDIT_ERR_INVAL;
			goto err;
		}

		while (i < 2) {
			int result;

			result = ogg_sync_pageout (&s->oy, &og);
			if (!result)
				break; /* Too little data so far */

			if (result != 1)
				continue;

			ogg_stream_pagein (&s->os, &og);

			while (i < 2) {
				result = ogg_stream_packetout (&s->os, header);
				if (!result)
					break;

				if (result != 1) {
					ret = VCEDIT_ERR_INVAL;
					goto err;
				}

				vorbis_synthesis_headerin (&s->vi, &s->vc, header);

				if (i++ == 1) {
					s->booklen = header->bytes;
					s->bookbuf = malloc (s->booklen);
					memcpy (s->bookbuf, header->packet, header->bytes);
				}

				header = &header_codebooks;
			}
		}

		buffer = ogg_sync_buffer (&s->oy, CHUNKSIZE);

		bytes = fread (buffer, 1, CHUNKSIZE, s->in);
		ogg_sync_wrote (&s->oy, bytes);
	}

	/* Copy the vendor tag */
	s->vendor = strdup (s->vc.vendor);

	/* Headers are done! */
	s->opened = true;

	return VCEDIT_ERR_SUCCESS;

err:
	vcedit_clear_internals (s);

	return ret;
}

static int
write_data (const void *buf, size_t size, size_t nmemb, FILE *stream)
{
	while (nmemb > 0) {
		size_t w;

		w = fwrite (buf, size, nmemb, stream);
		if (!w && ferror (stream))
			return 0;

		nmemb -= w;
		buf += size * w;
	}

	return 1;
}

vcedit_error
vcedit_write (vcedit_state *s)
{
	ogg_stream_state streamout;
	ogg_packet header_main, header_comments, header_codebooks, op;
	ogg_page ogout, ogin;
	ogg_int64_t granpos = 0;
	FILE *out;
	char *buffer, tmpfile[PATH_MAX];
	int fd, result, bytes, needflush = 0, needout = 0;

	if (!s->opened)
		return VCEDIT_ERR_INVAL;

	strcpy (tmpfile, s->filename);
	strcat (tmpfile, ".XXXXXX");

	fd = mkstemp (tmpfile);
	if (fd == -1)
		return VCEDIT_ERR_TMPFILE;

	out = fdopen (fd, "wb");
	if (!out) {
		unlink (tmpfile);
		close (fd);

		return VCEDIT_ERR_TMPFILE;
	}

	s->prevW = s->extrapage = s->eosin = 0;

	header_main.bytes = s->mainlen;
	header_main.packet = s->mainbuf;
	header_main.b_o_s = 1;
	header_main.e_o_s = 0;
	header_main.granulepos = 0;

	header_codebooks.bytes = s->booklen;
	header_codebooks.packet = s->bookbuf;
	header_codebooks.b_o_s = 0;
	header_codebooks.e_o_s = 0;
	header_codebooks.granulepos = 0;

	ogg_stream_init (&streamout, s->serial);

	_commentheader_out (&s->vc, s->vendor, &header_comments);

	ogg_stream_packetin (&streamout, &header_main);
	ogg_stream_packetin (&streamout, &header_comments);
	ogg_stream_packetin (&streamout, &header_codebooks);

	while ((result = ogg_stream_flush (&streamout, &ogout))) {
		if (!write_data (ogout.header, 1, ogout.header_len, out))
			goto cleanup;

		if (!write_data (ogout.body, 1, ogout.body_len, out))
			goto cleanup;
	}

	while (_fetch_next_packet (s, &op, &ogin)) {
		int size;

		size = _blocksize (s, &op);
		granpos += size;

		if (needflush) {
			if (ogg_stream_flush (&streamout, &ogout)) {
				if (!write_data (ogout.header, 1, ogout.header_len, out))
					goto cleanup;

				if (!write_data (ogout.body, 1, ogout.body_len, out))
					goto cleanup;
			}
		} else if (needout) {
			if (ogg_stream_pageout (&streamout, &ogout)) {
				if (!write_data (ogout.header, 1, ogout.header_len, out))
					goto cleanup;

				if (!write_data (ogout.body, 1, ogout.body_len, out))
					goto cleanup;
			}
		}

		needflush = needout = 0;

		if (op.granulepos == -1) {
			op.granulepos = granpos;
			ogg_stream_packetin (&streamout, &op);
		} else {
			/* granulepos is set, validly. Use it, and force a flush to
			 * account for shortened blocks (vcut) when appropriate
			 */
			if (granpos > op.granulepos) {
				granpos = op.granulepos;
				ogg_stream_packetin (&streamout, &op);
				needflush = 1;
			} else {
				ogg_stream_packetin (&streamout, &op);
				needout = 1;
			}
		}
	}

	streamout.e_o_s = 1;

	while (ogg_stream_flush (&streamout, &ogout)) {
		if (!write_data (ogout.header, 1, ogout.header_len, out))
			goto cleanup;

		if (!write_data (ogout.body, 1, ogout.body_len, out))
			goto cleanup;
	}

	if (s->extrapage) {
		if (!write_data (ogin.header, 1, ogin.header_len, out))
			goto cleanup;

		if (!write_data (ogin.body, 1, ogin.body_len, out))
			goto cleanup;
	}

	/* clear it, because not all paths to here do */
	s->eosin = 0;

	do {
		/* We copy the rest of the stream (other logical streams)
		 * through, a page at a time.
		 */
		while ((result = ogg_sync_pageout (&s->oy, &ogout))) {
			if (result != 1)
				continue;

			/* Don't bother going through the rest, we can just
			 * write the page out now
			 */
			if (!write_data (ogout.header, 1, ogout.header_len, out))
				goto cleanup;

			if (!write_data (ogout.body, 1, ogout.body_len, out))
				goto cleanup;
		}

		buffer = ogg_sync_buffer (&s->oy, CHUNKSIZE);
		bytes = fread (buffer, 1, CHUNKSIZE, s->in);
		ogg_sync_wrote (&s->oy, bytes);

		if (ferror (s->in))
			goto cleanup;

		s->eosin = !bytes && feof (s->in);
	} while (!s->eosin);

	fclose (out);
	fclose (s->in);

	unlink (s->filename);
	rename (tmpfile, s->filename);

cleanup:
	ogg_stream_clear (&streamout);

    /* We don't ogg_packet_clear() this, because the memory was
	 * allocated in _commentheader_out(), so we mirror that here
	 */
    _ogg_free (header_comments.packet);

	free (s->mainbuf);
	free (s->bookbuf);

    s->mainbuf = s->bookbuf = NULL;

	if (!s->eosin)
		return VCEDIT_ERR_INVAL;

	vcedit_clear_internals (s);

	return (vcedit_open (s) == VCEDIT_ERR_SUCCESS) ?
	       VCEDIT_ERR_SUCCESS : VCEDIT_ERR_REOPEN;
}
